unit u_regnt;

interface

uses
  Windows, SysUtils, Classes, Forms, FileCtrl, u_main;

function CopyHive(const Src, Dst: string; Options: TDCOptions): Boolean;
function ChangePrivilege(const Privilege: string; const Attr: Integer): boolean;
function RegisterEventLogApp: Boolean;
function LogNTEvent(Message: string; const eType: TLogEventType): Boolean;

implementation

uses u_str;

type
  TCheckHive = (chNone, chBackup, chRestore);

var
  HiveList: TStringList;

function DeviceToDrive(dStr: string): string;
var
  s: set of 0..25;
  i: Integer;
  sr: array[0..Max_Path] of Char;
  sr1: string;
begin
  Result := AnsiUpperCase(dStr);
  Integer(s) := GetLogicalDrives;
  for i := 0 to 25 do
  begin
    if i in s then
    begin
      sr1 := Char(i + Ord('A')) + ':';
      QueryDosDevice(PChar(sr1), sr, Max_Path);
      sr1 := AnsiUpperCase(StrPas(sr));
      if Pos(sr1, Result) = 1 then
      begin
        Delete(Result, 1, Length(sr1));
        if (Result <> '') and (Result[1] = '\')
          then
          Delete(Result, 1, 1);
        Result := Char(i + Ord('A')) + ':\' + Result;
        Break;
      end;
    end;
  end;
end;

function ReplaceRegStr(rStr: string): string;
begin
  Result := AnsiUpperCase(rStr);
  if Pos('\REGISTRY\MACHINE\', Result) = 1 then
  begin
    Delete(Result, 1, 18);
    Insert('$' + IntToHex(HKEY_LOCAL_MACHINE, 8) + ',', Result, 1);
  end
  else
    if Pos('\REGISTRY\USER\', Result) = 1 then
    begin
      Delete(Result, 1, 15);
      Insert('$' + IntToHex(HKEY_USERS, 8) + ',', Result, 1);
    end
end;

function ChangePrivilege(const Privilege: string; const Attr: Integer): boolean;
var
  hToken: THandle;
  ptkp, ptkpold: PTokenPrivileges;
  r: DWORD;
  le: DWORD;
const
  ANYSIZE_ARRAY = 1;
begin
  Result := FALSE;
  le := GetLastError;
  try
    if not IsNT
      then
      Result := True
    else
    begin
      if OpenProcessToken(GetCurrentProcess, TOKEN_ADJUST_PRIVILEGES or TOKEN_QUERY, hToken) then
      begin
        GetMem(ptkp, SizeOf(TTOKENPRIVILEGES) + (ANYSIZE_ARRAY) *
          SizeOf(TLUIDANDATTRIBUTES));
        LookupPrivilegeValue(nil, PChar(Privilege), ptkp^.Privileges[0].Luid);
        ptkp^.PrivilegeCount := 1;
        ptkp^.Privileges[0].Attributes := Attr;
        r := 0;
        ptkpold := nil;
        Result := AdjustTokenPrivileges(hToken, FALSE, ptkp^, 0, ptkpold^, r);
        FreeMem(ptkp);
      end;
    end;
  finally
    if Result
      then
      SetLastError(le);
  end;
end;

function CheckHive(const FNameSrc, FNameDst: string; var HiveIndex: Integer): TCheckHive;
var
  i: Integer;
  sr, srs, srd: string;
begin
  HiveIndex := -1;
  Result := chNone;
  if (not IsNT) or (HiveList.Count < 1)
    then
    Exit
  else
  begin
    srs := AnsiUpperCase(FNameSrc);
    srd := AnsiUpperCase(FNameDst);
    for i := 0 to Pred(HiveList.Count) do
    begin
      sr := AnsiUpperCase(HiveList.Values[HiveList.Names[i]]);
      if sr = srs then
      begin
        HiveIndex := i;
        Result := chBackup;
        Break;
      end
      else
        if sr = srd then
        begin
          HiveIndex := i;
          Result := chRestore;
          Break;
        end
    end;
  end;
end;

function SaveReg(const Path: string; const HiveIndex: Integer): Boolean;
var
  rc: Integer;
  rKey: HKey;
  dwDisposition: DWORD;
  root: HKey;
  Key: string;
begin
  Result := False;
  if not ChangePrivilege('SeBackupPrivilege', SE_PRIVILEGE_ENABLED)
    then
    Exit;
  try
    Key := HiveList.Names[HiveIndex];
    root := StrToInt(Copy(Key, 1, 9));
    delete(Key, 1, 10);
    rc := RegCreateKeyEx(root, PChar(key), 0, nil,
      REG_OPTION_BACKUP_RESTORE, KEY_READ, nil, rKey, @dwDisposition);

    if rc = ERROR_SUCCESS then
    begin
      DelFile(Path, True);
      rc := RegSaveKey(rKey, PChar(Path), nil);
      Result := rc = ERROR_SUCCESS;
      if not Result
        then
        SetLastError(rc);
      RegCloseKey(rKey);
    end
    else
      SetLastError(rc);

  finally
    ChangePrivilege('SeBackupPrivilege', 0);
  end;
end;

procedure RegLoadError(const mess: string);
begin
  SetForeGroundWindow(Application.Handle);
  MessageBox(0, PChar(mess), PChar(rleErr), MB_ICONERROR);
end;

function LoadReg(const SPath, TPath: string; const HiveIndex: Integer): Boolean;
var
  rc: Integer;
  root: HKey;
  Key, sname, bname: string;
  rname: string;
begin
  Result := False;
  if not ChangePrivilege('SeRestorePrivilege', SE_PRIVILEGE_ENABLED)
    then
    Exit;
  try
    Key := HiveList.Names[HiveIndex];
    root := StrToInt(Copy(Key, 1, 9));
    case root of
      HKEY_LOCAL_MACHINE: rname := 'HKLM\';
      HKEY_USERS: rname := 'HKU\';
    else
      rname := '??\';
    end;
    delete(Key, 1, 10);

    // abfragen
    SetForeGroundWindow(Application.Handle);
    if MessageBox(0, PChar(Format(rmovreg, [rname + Key, SPath])), PChar(rmovwa), MB_ICONEXCLAMATION or MB_YESNO) <> idYes
      then
      Exit;

    // copy the new hive into the target directory
    sname := ChangeFileExt(TPath, '.drh');
    if FileExists(sname) then
    begin
      if not CopyFile(PChar(sname), PChar(sname + '.000'), False) then
      begin
        RegLoadError(Format(rlecpb, [sname, sname + '.000']));
        Exit;
      end;
      DelFile(sname, True);
    end;

    if not CopyFile(PChar(SPath), PChar(sname), False) then
    begin
      RegLoadError(Format(rlecph, [SPath, sname]));
      Exit;
    end;

    bname := ChangeFileExt(TPath, '.dbh');
    if FileExists(bname) then
    begin
      if not CopyFile(PChar(bname), PChar(bname + '.000'), False) then
      begin
        RegLoadError(Format(rlecpb, [bname, bname + '.000']));
        Exit;
      end;
      DelFile(bname, True);
    end;

    rc := RegReplaceKey(root, PChar(Key), PChar(sname), PChar(bname));
    Result := rc = ERROR_SUCCESS;
    if not Result then
    begin
      SetLastError(rc);
      RegLoadError(Format(rlecpr, [rname + Key, sname, bname]));
    end
    else
    begin
      SetForeGroundWindow(Application.Handle);
      MessageBox(0, PChar(Format(rlecps, [rname + Key, sname, bname])), PChar(rleSucc), MB_ICONINFORMATION);
    end;

  finally
    ChangePrivilege('SeRestorePrivilege', 0);
  end;
end;

function LookUpHives: Boolean;
var
  rc: LongInt;
  rKey: HKey;
  dwDisposition: DWORD;
  j: Integer;
  ValName, ValValue: array[0..Max_Path] of Char;
  LenName, LenValue: Cardinal;
  ValType: Integer;
begin
  HiveList.Clear;

  // not possible / necessary on win 9x, so return true
  if not IsNT then
  begin
    Result := True;
    Exit;
  end;
  Result := False;

  // Get Backup rights, hopefully can read it then
  if not ChangePrivilege('SeBackupPrivilege', SE_PRIVILEGE_ENABLED)
    then
    Exit;

  // open the key that contains the currently loaded hive's list
  rc := RegCreateKeyEx(HKEY_LOCAL_MACHINE, 'SYSTEM\CURRENTCONTROLSET\CONTROL\HIVELIST', 0, nil,
    REG_OPTION_BACKUP_RESTORE, KEY_READ, nil, rKey, @dwDisposition);

  // couldn't open the list
  if rc <> ERROR_SUCCESS then
  begin
    ChangePrivilege('SeBackupPrivilege', 0);
    Exit;
  end;
  try
    // enumerate the list
    j := 0;
    repeat
      LenName := Max_Path;
      LenValue := Max_Path;
      rc := RegEnumValue(rKey, j, ValName, LenName, nil, @ValType, @ValValue[0], @LenValue);
      case rc of
        ERROR_SUCCESS:
          begin
            case ValType of
              REG_SZ, REG_EXPAND_SZ, REG_MULTI_SZ:
                if StrPas(ValValue) <> ''
                  then
                  HiveList.Add(ReplaceRegStr(StrPas(ValName)) + '=' + DeviceToDrive(StrPas(ValValue)));
            end;
            Inc(J);
            Result := True;
          end;
        ERROR_NO_MORE_ITEMS: Break;
      else
        Result := False;
        HiveList.Clear;
        Break;
      end;
    until False;

  finally
    RegCloseKey(rKey);
    ChangePrivilege('SeBackupPrivilege', 0);
  end;
end;

function CopyHive(const Src, Dst: string; Options: TDCOptions): Boolean;
var
  i: Integer;
  ch: TCheckHive;
begin
  Result := False;
  if not LookUpHives
    then
    Exit;

  ch := CheckHive(src, dst, i);

  if (ch <> chNone) and (not Options.oForceReg) then
  begin
    SetLastError(SE_NOFORCEREG);
    Exit;
  end;

  case ch of
    chBackup: Result := SaveReg(Dst, i);
    chRestore:
      if not Opts.oAuto
        then
        Result := LoadReg(Src, Dst, i)
      else
        SetLastError(SE_NOAUTORESTOREREG);
  end;
end;

function RegisterEventLogApp: Boolean;
var
  hk: HKEY;
  dwData: DWORD;
  szBuf: string;
  rs: Cardinal;
begin
  Result := False;
  if not IsNT
    then
    Exit;

  rs := RegCreateKey(HKEY_LOCAL_MACHINE, PChar('SYSTEM\CurrentControlSet\Services\EventLog\Application\' + aeName), hk);
  if rs <> ERROR_SUCCESS then
  begin
    SetLastError(rs);
    Exit;
  end;

  // Set the Event ID message-file name. (its dcopy.dll, the second project)

  szBuf := ChangeFileExt(ParamStr(0), '.dll') + #0;


  // Add the Event ID message-file name to the subkey.

  rs := RegSetValueEx(hk, 'EventMessageFile', 0, REG_EXPAND_SZ, @szBuf[1], Length(szBuf));
  if rs <> ERROR_SUCCESS then
  begin
    SetLastError(rs);
    RegCloseKey(hk);
    Exit;
  end;


  // Set the supported types flags.

  dwData := EVENTLOG_ERROR_TYPE or EVENTLOG_WARNING_TYPE or EVENTLOG_INFORMATION_TYPE or EVENTLOG_AUDIT_SUCCESS;

  rs := RegSetValueEx(hk, 'TypesSupported', 0, REG_DWORD, @dwData, SizeOf(DWORD));
  if rs <> ERROR_SUCCESS then
  begin
    SetLastError(rs);
    RegCloseKey(hk);
    Exit;
  end;

  RegCloseKey(hk);
  Result := True;
end;

const
  EVENT_EXCEPTION = $C00004B0;
  EVENT_INFORMATION = $400004B1;
  EVENT_WARNING = $800004B2;
  EVENT_ERROR = $C00004B3;
  EVENT_SUCCESS = $000004B4;

  LogTypes: array[TLogEventType] of Cardinal = (EVENTLOG_ERROR_TYPE, EVENTLOG_INFORMATION_TYPE,
    EVENTLOG_INFORMATION_TYPE,
    EVENTLOG_INFORMATION_TYPE);

  EventTypes: array[TLogEventType] of Cardinal = (EVENT_ERROR, EVENT_INFORMATION, EVENT_INFORMATION, EVENT_SUCCESS);

function LogNTEvent(Message: string; const eType: TLogEventType): Boolean;
var
  h: THandle;
  msg: PChar;
  et, lt: Cardinal;
begin
  Result := False;
  h := RegisterEventSource(nil, aeName);
  if h = 0
    then
    Exit;

  et := EventTypes[eType];
  lt := LogTypes[eType];

  Message := Trim(Message);

  if Pos('w#', Message) = 1 then
  begin
    Delete(Message, 1, 2);
    et := EVENT_WARNING;
    lt := EVENTLOG_WARNING_TYPE;
  end
  else
    if Pos('e#', Message) = 1 then
    begin
      Delete(Message, 1, 2);
      et := EVENT_ERROR;
      lt := EVENTLOG_ERROR_TYPE;
    end;

  msg := PChar(Message);

  Result := ReportEvent(h, lt, 0, et, nil, 1, 0, @msg, nil);
  DeregisterEventSource(h);
end;

initialization
  HiveList := TStringList.Create;
finalization
  HiveList.Free;

end.
